module EncodingLinear1 where

import Data.List
import Term
import Rule
import TRS
import SMT
import Polynomial

-- f_0 + f_1 x_1 + ... + f_n x_n
type Coefficients = (Int, [Int])

-- f(x1,...,xn) = f_0 + f_1 x_1 + ... + f_n x_n
type Interpretation = (String,Coefficients)

type Algebra = [Interpretation]

type Proof = [(Algebra,TRS)]

showMonomial :: (Int, String) -> String
showMonomial (0, _) = "0"
showMonomial (1, x) = x
showMonomial (n, x) = show n ++ " " ++ x

-- print as polynomials
showCoefficients :: (String, Coefficients) -> String
showCoefficients (x, (a0, as)) = 
  case ms1 ++ ms2 of
    [] -> "0"
    ms -> intercalate " + " ms
  where
    ms1 = [ showMonomial (ai, x ++ show i)
          | (i, ai) <- zip [1 :: Int ..] as, ai /= 0 ]
    ms2 = if a0 == 0 then [] else [show a0]

{-
showInterpretation :: (String, Interpretation) -> String
showInterpretation (s, (f, (a0,[]))) =
  f ++ "_" ++ s ++ " = " ++ show a0
showInterpretation (s, (f, p@(_,as))) =
  f ++ "_" ++ s ++ "(" ++ xs ++ ") = " ++ showCoefficients p
  where
    xs = intercalate "," [ "x" ++ show i | i <- [1 .. length as] ]
-}

showInterpretation :: (String, Coefficients) -> String
showInterpretation (f, (a0,[])) =
  f ++ "_A" ++ " = " ++ show a0
showInterpretation (f, p@(_,as)) =
  f ++ "_A" ++ "(" ++ xs ++ ") = " ++ showCoefficients ("x",p)
  where
    xs = intercalate "," [ "x" ++ show i | i <- [1 .. length as] ]

showAlgebra :: Algebra -> String
showAlgebra interpretationsA =
  unlines
    (["The weakly monotone algebra (N, >) with",
      ""] ++
     ["  " ++ showInterpretation (f, p) 
     | (f, p) <- interpretationsA])

interpret :: Algebra -> Term -> Polynomial
interpret _ (V x) = var x 
interpret a (F f ts)
  | Just (a0, as) <- lookup f a = 
    Polynomial.plus (Polynomial.num a0 : [ Polynomial.times ai (interpret a ti) | (ai, ti) <- zip as ts ])
interpret _ _ = error "interpret"

-- >=_A
geqA :: Algebra -> Rule -> Bool
geqA a (l, r) = Polynomial.geq (interpret a l) (interpret a r)

-- >_A
gtA :: Algebra -> Rule -> Bool
gtA a (l, r) = Polynomial.gt (interpret a l) (interpret a r)


-- Encoding functions.

varF0 :: Int -> String -> Exp
varF0 k f = Var ("a0_" ++ show k ++ "_" ++ f) 

varFi :: Int -> String -> Int -> Formula
varFi k f i = FVar ("a_" ++ show k ++ "_" ++ show i ++ "_" ++ f) 

constantPart :: Int -> Term -> Exp
constantPart _ (V _)    = Val 0
constantPart k (F f ts) =
  SMT.plus (varF0 k f : [ times01 (varFi k f i) (constantPart k ti)
                        | (i,ti) <- zip [1..] ts ])

coefficient :: Int -> String -> Term -> Exp
coefficient _ x (V y)
  | x == y    = Val 1
  | otherwise = Val 0
coefficient k x (F f ts) =
  SMT.plus [ times01 (varFi k f i) (EncodingLinear1.coefficient k x ti) | (i, ti) <- zip [1..] ts ]

-- s >=_A t
encodeGeqA :: Int -> Rule -> Formula
encodeGeqA k rule@(l, r) =
  conj (SMT.geq (constantPart k l) (constantPart k r) :
        [ SMT.geq (EncodingLinear1.coefficient k x l)
                  (EncodingLinear1.coefficient k x r) 
        | x <- Rule.variables rule ])

-- s >_A t
encodeGtA :: Int -> Rule -> Formula
encodeGtA k rule@(l, r) =
  conj (SMT.gt (constantPart k l) (constantPart k r) :
        [ SMT.geq (EncodingLinear1.coefficient k x l)
                  (EncodingLinear1.coefficient k x r) 
        | x <- Rule.variables rule ])

encode :: TRS -> TRS -> Formula
encode trs1 trs2 = 
  conj ([ SMT.geq (varF0 1 f) (Val 0) 
        | (f, _) <- sig ] ++
        [ encodeGeqA 1 rule | rule <- trs1 ++ trs2 ] ++
        [ disj [ encodeGtA 1 rule | rule <- trs1 ] ])
  where
    sig = TRS.signatureOf (trs1 ++ trs2)

-- Decoding functions

decodeInterpretation :: Int -> String -> Int -> Model -> Interpretation
decodeInterpretation k f n model = (f, (f0, fs))
  where 
    f0 = evalExp model (varF0 k f)
    fs = [ if evalFormula model (varFi k f i) then 1 else 0 | i <- [1 .. n] ]

decodeAlgebra :: TRS -> Model -> Int -> Algebra
decodeAlgebra trs model k =
  [ decodeInterpretation k f n model 
  | (f, n) <- TRS.signatureOf trs ]

solveOne :: String -> TRS -> TRS -> IO (Maybe (Algebra, TRS))
solveOne tool trs1 trs2 = do
  result <- sat tool (encode trs1 trs2)
  case result of
    Nothing           -> return Nothing
    Just model 
      | removed == [] -> error "solveOne"
      | otherwise     -> return (Just (a, removed))
      where 
        a = decodeAlgebra (trs1 ++ trs2) model 1
        removed = [ rule | rule <- trs1, gtA a rule ]
