module EncodingMaxPlus where

import Data.List
import Term
import Rule
import TRS
import SMT
import MaxPlus
import SymbolicMaxPlus as SMaxPlus

-- (f0, [(f1, f'1), ..., (fn, f'n)]) represents
-- max{ f0, f1 + f'1 x_1, ..., fn + f'n x_n }
type MaxPlusParam = (Int, [(Int, Int)])

-- f(x1,...,xn) = max{ f0, f1 + f'1 x_1, ..., fn + f'n x_n }
type Interpretation = (String, MaxPlusParam)

type Algebra = [Interpretation]

type Proof = [(Algebra,TRS)]

showMonomial :: (Int, String) -> String
showMonomial (0, _) = "0"
showMonomial (1, x) = x
showMonomial (n, x) = show n ++ " " ++ x

-- pretty printer for f + f'x
showLinearPolynomial :: (Int, Int, String) -> String
showLinearPolynomial (f, 0, _x) = show f
showLinearPolynomial (0, 1, x) = x
showLinearPolynomial (f, 1, x) = show f ++ " + " ++ x
showLinearPolynomial (0, f', x) = show f' ++ x
showLinearPolynomial (f, f', x) = show f ++ " + " ++ show f' ++ " " ++ x

-- can be improved
showMaxPlus :: MaxPlusParam -> String
showMaxPlus (a0, []) = show a0
showMaxPlus (a0, as) = "max{" ++ (intercalate ", " ms) ++ "}"
  where
    m1n = [ showLinearPolynomial (ai, ai', "x" ++ show i)
          | (i, (ai, ai')) <- zip [ 1 :: Int ..] as]
    ms = show a0 : m1n

showInterpretation :: Interpretation -> String
showInterpretation (f, (a0, [])) =
  f ++ "_A = " ++ show a0
showInterpretation (f, mp@(_, as)) =
  f ++ "_A(" ++ xs ++ ") = " ++ showMaxPlus mp
  where
    xs = intercalate "," [ "x" ++ show i | i <- [1 .. length as] ]

showAlgebra :: Algebra -> String
showAlgebra interpretations =
  "max/plus interpretations on N:\n" ++
  -- show interpretations ++ -- DEBUG
  "\n" ++
  unlines 
    [ "  " ++ showInterpretation interpretation 
    | interpretation <- interpretations ] ++ "\n"

showProof :: Proof -> String
showProof [] = "We are done.\n"
showProof ((a,removed) : pairs) =
  showAlgebra a ++
  "\nWe remove the strictly oriented rule(s):\n\n" ++
  showTRS removed ++ "\n" ++
  "--\n\n" ++
  showProof pairs

interpret :: Algebra -> Term -> MaxPlus
interpret _ (V x) = MaxPlus.var x 
interpret a (F f ts)
  | Just (a0, as) <- lookup f a = 
    MaxPlus.flatten (MaxPlus.num a0 : [ MaxPlus.plus ai (MaxPlus.times ai' (interpret a ti)) | ((ai, ai'), ti) <- zip as ts ])
interpret _ _ = error "interpret"

-- >=_A
geqA :: Algebra -> Rule -> Bool
geqA a (l, r) = MaxPlus.geq (interpret a l) (interpret a r)

-- >_A
gtA :: Algebra -> Rule -> Bool
gtA a (l, r) = MaxPlus.gt (interpret a l) (interpret a r)

-- Encoding functions.

varF0 :: String -> Exp
varF0 f = Var ("a0_" ++ f) 

varFi :: String -> Int -> Exp
varFi f i = Var ("a_" ++ show i ++ "_" ++ f) 

varF'i :: String -> Int -> Formula
varF'i f i = FVar ("b_" ++ show i ++ "_" ++ f) 

symbolicMaxPlus :: Term -> SymbolicMaxPlus
symbolicMaxPlus (V x) = SMaxPlus.var x
symbolicMaxPlus (F f ts) =
   SMaxPlus.flatten (SMaxPlus.const (varF0 f) : [ multiplyAdd f i ti | (i, ti) <- zip [1..] ts ])
  where
    multiplyAdd fname i ti = SMaxPlus.addConstant (varFi fname i) (SMaxPlus.mult (varF'i fname i) (symbolicMaxPlus ti))

-- s >=_A t
encodeGeqA :: Rule -> Formula
encodeGeqA (l, r) = SMaxPlus.encodeGeq (symbolicMaxPlus l) (symbolicMaxPlus r)

-- s >_A t
encodeGtA :: Rule -> Formula
encodeGtA (l, r) = SMaxPlus.encodeGt (symbolicMaxPlus l) (symbolicMaxPlus r)

encode :: TRS -> TRS -> Formula
encode trs1 trs2 = 
  conj ([ SMT.geq (varF0 f) (Val 0) | (f, _) <- sig ] ++
        [ encodeGeqA rule | rule <- trs1 ++ trs2 ] ++
        [ disj [ encodeGtA rule | rule <- trs1 ] ])
  where
    sig = TRS.signatureOf (trs1 ++ trs2)

-- decoding

decodeInterpretation :: String -> Int -> Model -> Interpretation
decodeInterpretation f n model = (f, (f0, ff's))
  where 
    f0 = evalExp model (varF0 f)
    evalf fname i = evalExp model (varFi fname i)
    evalf' fname i = if evalFormula model (varF'i fname i) then 1 else 0
    ff's = [ (evalf f i, evalf' f i) | i <- [1 .. n] ]

decodeAlgebra :: TRS -> Model -> Algebra
decodeAlgebra trs model =
  [ decodeInterpretation f n model | (f, n) <- TRS.signatureOf trs ]

solveOne :: String -> TRS -> TRS -> IO (Maybe (Algebra, TRS))
solveOne tool trs1 trs2 = do
  result <- sat tool (encode trs1 trs2)
  case result of
    Nothing           -> return Nothing
    Just model 
      | removed == [] -> error "solveOne"
      | otherwise     -> return (Just (a, removed))
      where 
        a = decodeAlgebra (trs1 ++ trs2) model
        removed = [ rule | rule <- trs1, gtA a rule ]

solve :: String -> TRS -> TRS -> IO (Maybe [(Algebra,TRS)])
solve _    []   _    = return (Just []) 
solve tool trs1 trs2 = do
  result1 <- solveOne tool trs1 trs2
  case result1 of
    Nothing              -> return Nothing
    Just (pair, removed) -> do
      let trs1' = [ rule | rule <- trs1, notElem rule removed ]  
      result2 <- solve tool trs1' trs2
      case result2 of
        Nothing    -> return Nothing
        Just pairs -> return (Just ((pair, removed) : pairs))