module EncodingPolynomial where

import Data.List
import Term
import Rule
import TRS
import SMT
import Polynomial

-- f_0 + f_1 x_1 + ... + f_n x_n
type Coefficients = (Int, [Int])

-- f(x1,...,xn) = f_0 + f_1 x_1 + ... + f_n x_n
type Interpretation = (String,Coefficients)

type Algebra = [Interpretation]

showMonomial :: (Int, String) -> String
showMonomial (0, _) = "0"
showMonomial (1, x) = x
showMonomial (n, x) = show n ++ " " ++ x

-- print as polynomials
showCoefficients :: Coefficients -> String
showCoefficients (a0, as) = 
  case ms1 ++ ms2 of
    [] -> "0"
    ms -> intercalate " + " ms
  where
    ms1 = [ showMonomial (ai, "x" ++ show i) 
          | (i, ai) <- zip [1 :: Int ..] as, ai /= 0 ]
    ms2 = if a0 == 0 then [] else [show a0]

showInterpretation :: Interpretation -> String
showInterpretation (f, (a0,[])) =
  f ++ "_A = " ++ show a0
showInterpretation (f, p@(_,as)) =
  f ++ "_A(" ++ xs ++ ") = " ++ showCoefficients p
  where
    xs = intercalate "," [ "x" ++ show i | i <- [1 .. length as] ]

showAlgebra :: Algebra -> String
showAlgebra interpretations =
  "linear polynomial interpretations on N:\n" ++
  "\n" ++
  unlines 
    [ "  " ++ showInterpretation interpretation 
    | interpretation <- interpretations ] ++ "\n"

interpret :: Algebra -> Term -> Polynomial
interpret _ (V x) = var x 
interpret a (F f ts)
  | Just (a0, as) <- lookup f a = 
    Polynomial.plus (Polynomial.num a0 : [ Polynomial.times ai (interpret a ti) | (ai, ti) <- zip as ts ])
interpret _ _ = error "interpret"

-- >=_A
geqA :: Algebra -> Rule -> Bool
geqA a (l, r) = Polynomial.geq (interpret a l) (interpret a r)

-- >_A
gtA :: Algebra -> Rule -> Bool
gtA a (l, r) = Polynomial.gt (interpret a l) (interpret a r)

-- Encoding functions.

varF0 :: String -> Exp
varF0 f = Var ("a0_" ++ f) 

varFi :: String -> Int -> Formula
varFi f i = FVar ("a_" ++ show i ++ "_" ++ f) 

constantPart :: Term -> Exp
constantPart (V _)    = Val 0
constantPart (F f ts) =
  SMT.plus (varF0 f : [ times01 (varFi f i) (constantPart ti)
                      | (i,ti) <- zip [1..] ts ])

coefficient :: String -> Term -> Exp
coefficient x (V y)
  | x == y    = Val 1
  | otherwise = Val 0
coefficient x (F f ts) =
  SMT.plus [ times01 (varFi f i) (EncodingPolynomial.coefficient x ti) | (i, ti) <- zip [1..] ts ]

-- s >=_A t
encodeGeqA :: Rule -> Formula
encodeGeqA rule@(l, r) =
  conj (SMT.geq (constantPart l) (constantPart r) :
        [ SMT.geq (EncodingPolynomial.coefficient x l) (EncodingPolynomial.coefficient x r) 
        | x <- Rule.variables rule ])

-- s >_A t
encodeGtA :: Rule -> Formula
encodeGtA rule@(l, r) =
  conj (SMT.gt (constantPart l) (constantPart r) :
        [ SMT.geq (EncodingPolynomial.coefficient x l) (EncodingPolynomial.coefficient x r) 
        | x <- Rule.variables rule ])

encode :: TRS -> TRS -> Formula
encode trs1 trs2 = conj (map encodeGtA trs1 ++ map encodeGeqA trs2 ++ nonNeg)
  where
    nonNeg = [ SMT.geq (varF0 f) (Val 0) | (f, _) <- TRS.signatureOf (trs1 ++ trs2) ]

-- Decoding functions

decodeInterpretation :: String -> Int -> Model -> Interpretation
decodeInterpretation f n model = (f, (f0, fs))
  where 
    f0 = evalExp model (varF0 f)
    fs = [ if evalFormula model (varFi f i) then 1 else 0 | i <- [1 .. n] ]

decodeAlgebra :: TRS -> Model -> Algebra
decodeAlgebra trs model =
  [ decodeInterpretation f n model 
  | (f, n) <- TRS.signatureOf trs ]

solve :: String -> TRS -> TRS -> IO (Maybe Algebra)
solve tool trs1 trs2 = do
  result <- sat tool (encode trs1 trs2)
  case result of
    Nothing    -> return Nothing
    Just model -> do
      let a = decodeAlgebra (trs1 ++ trs2) model
      return (Just a)
