module Main where

import System.Environment
import Rule
import TRS
import Parser
import Recursive
import RelativeDP

dominance :: TRS -> TRS -> Bool
dominance trs1 trs2 =
  dominates trs1 trs2 && TRS.non_duplicating trs2

almostDominance :: TRS -> TRS -> Bool
almostDominance trs1 trs2 =
  almostDominates trs1 trs2 && TRS.non_duplicating trs2

printProof :: (TRS, TRS, TRS, TRS, Proof) -> IO ()
printProof (trs1, trs2, ps, qs, proof) = do
  putStrLn "YES"
  putStrLn ""
  putStrLn "We show the termination of R/S, where R is"
  putStrLn ""
  putStr (showTRS trs1)
  putStrLn ""
  putStrLn "and S is:"
  putStrLn ""
  putStr (showTRS trs2)
  putStrLn ""
  putStrLn "Since R almost dominates S and S is non-duplicating,"
  putStrLn "it is enough to show finiteness of (P, Q)."
  putStrLn "Here P consists of the dependency pairs"
  putStrLn ""
  putStr (showTRS ps)
  putStrLn ""
  putStrLn "and Q consists of the rules:"
  putStrLn ""
  putStr (showTRS qs)
  putStrLn ""
  putStr (showProof proof)


prove :: [Criterion] -> TRS -> TRS -> IO (Maybe (TRS,TRS,TRS,TRS,Proof))
prove criteria trs1 trs2
  | almostDominance trs1 trs2 =
    let 
      ds = definedSymbols trs1
      ps = dp ds trs1
      qs = trs1 ++ trs2 ++ dp ds trs2
    in do
      result <- Recursive.solve criteria ps qs
      case result of
        Nothing    -> return Nothing
        Just proof -> return (Just (trs1,trs2,ps,qs,proof))
  | otherwise = return Nothing        

run :: [Criterion] -> TRS -> TRS -> IO ()
run criteria trs1 trs2 = do
  result <- prove criteria trs1 trs2
  case result of
    Nothing -> putStrLn "MAYBE"
    Just tuple -> printProof tuple

powerset :: [a] -> [[a]]
powerset [] = [[]]
powerset (x : xs) = yss ++ [ x : ys | ys <- yss ]
  where yss = powerset xs

loop :: [Criterion] -> TRS -> TRS -> [TRS] -> IO ()
loop _ _ _ [] = putStrLn "MAYBE"
loop criteria trs1 trs2 (trs : trss) = do
  result <- prove criteria trs1' trs2'
  case result of
    Nothing    -> loop criteria trs1 trs2 trss
    Just tuple -> printProof tuple
  where
    trs1' = trs1 ++ trs
    trs2' = [ rule | rule <- trs2, notElem rule trs ]


run2 :: [Criterion] -> TRS -> TRS -> IO ()
run2 criteria trs1 trs2 = 
  loop criteria trs1' trs2' (powerset trs2')
  where
    trs1' = [ rule | rule <- trs1 ] ++
            [ rule | rule <- trs2, not (Rule.non_duplicating rule) ]
    trs2' = [ rule | rule <- trs2, Rule.non_duplicating rule ]

check :: (TRS -> TRS -> Bool) -> TRS -> TRS -> IO ()
check applicable trs1 trs2
  | applicable trs1 trs2 = putStrLn "YES"
  | otherwise            = putStrLn "NO"

runHelp :: IO ()
runHelp = do
  putStrLn "A relative termination tool verion 0.11."
  putStrLn "Usage: tool [option] <file>"
  putStrLn "  where [file] is a two-TRS file"
  putStrLn "Options:"
  putStrLn "  -l1  linear interpretations on N equipped with >_lex:"
  putStrLn "         f_A(x_1,...,x_n) = f_0 + f_1 x_1 + ... + f_n x_n"
  putStrLn "         where f_0 in N, f_1,...,f_n in {0,1}"
  putStrLn "  -l2  linear interpretations on N^2 equipped with >_lex:"
  putStrLn "         f_A((x_1,y_1), ..., (x_n,y_n)) ="
  putStrLn "           (f_0 + f_1 x_1 + ... + f_n x_n, f'_0 + f'_1 y_1 + ... + f'_n y_n)"
  putStrLn "         where f_0, f'_0 in N, f_1,...,f_n,f'_1,...,f'_n in {0,1}"
  putStrLn "  -l3  linear interpretations on N^3 equipped with >_lex"
{-
  putStrLn "  -m   max/plus interpretations on N:"
  putStrLn "         f_A(x_1, ..., x_n) = max{f_0, f_1 x_1 + c_1, ..., f_n x_n + c_n}"
  putStrLn "         where f_0 in N, f_1,...,f_n in {0,1}, and c_1,...,c_n in Z"
  putStrLn "  -lm  combination of -o and -m"
-}
  putStrLn "  -d   checks if, given R/S, S is non-duplicating and R dominates S"
  putStrLn "  -a   checks if, given R/S, S is non-duplicating and R almost dominates S"

smt :: String
smt = "z3"

command :: String -> TRS -> TRS -> IO ()
command "-d"  = check dominance
command "-a"  = check almostDominance
command "-L"  = run  [linearAll smt]
command "-l1" = run  [linear1 smt]
command "-l2" = run  [linear2 smt]
command "-l3" = run  [linear3 smt]
command "-m"  = run  [maxplus smt]
command "-lm" = run  [linear3 smt, maxplus smt]
command _     = error "unknown option name"

read_and_run :: String -> String -> IO ()
read_and_run option filename = do
  errorOrTRS <- readTRSFile filename
  case errorOrTRS of
    Left e    -> putStrLn (show e)
    Right (trs1, trs2) -> 
      if not (TRS.variableCondition (trs1 ++ trs2)) then do
        putStrLn "ERROR"
        putStrLn ""
        putStrLn "The input TRS(s) violates the variable condition."
      else
        command option trs1 trs2

main :: IO ()
main = do
  args <- getArgs
  case args of
    [option, file] -> read_and_run option file
    [file]         -> read_and_run "-l3"  file
    _              -> runHelp
